import os
from collections.abc import Iterable

from memorious.logic.manager import Crawler, CrawlerManager


class TestManager(object):
    def test_manager(self):
        file_path = os.path.realpath(__file__)
        crawler_dir = os.path.normpath(os.path.join(
            file_path, "../testdata/config"
        ))
        manager = CrawlerManager()
        assert len(manager) == 0
        manager.load_path(crawler_dir)
        assert isinstance(manager.crawlers, dict)
        assert all(
            isinstance(crawler, Crawler) for crawler in manager
        )
        assert len(manager) == 3
        assert isinstance(manager.get("book_scraper"), Crawler)
        assert isinstance(manager["book_scraper"], Crawler)
        assert isinstance(manager, Iterable)



import locale
import logging
import threading

from memorious.core import settings

LANGUAGES = 'eng'
log = logging.getLogger(__name__)
lock = threading.RLock()


def get_ocr():
    """Check if OCR service is available; else throw an error"""
    if not hasattr(settings, '_ocr'):
        try:
            from tesserocr import PyTessBaseAPI, PSM, OEM
            log.info("Configuring OCR engine...")
            settings._ocr = PyTessBaseAPI(lang=LANGUAGES,
                                          oem=OEM.LSTM_ONLY,
                                          psm=PSM.AUTO_OSD)
        except ImportError:
            log.warning("OCR engine is not available")
            settings._ocr = None
    return settings._ocr


def read_text(image, languages=None):
    """OCR text from an image."""
    with lock:
        api = None
        currlocale = locale.getlocale()
        try:
            locale.setlocale(locale.LC_CTYPE, 'C')
            api = get_ocr()
            if api is None or image is None:
                return ''
            api.SetImage(image)
            return api.GetUTF8Text() or ''
        finally:
            if api is not None:
                api.Clear()
            locale.setlocale(locale.LC_CTYPE, currlocale)


def read_word(image, languages=None, spaces=False):
    """ OCR a word from an image. Useful for captchas.
        Image should be pre-processed to remove noise etc. """
    guess = read_text(image, languages=languages)
    if not spaces:
        guess = ''.join([c for c in guess if c != " "])
        guess = guess.strip()
    return guess


def read_char(image, languages=None):
    return read_word(image, languages=languages)


import logging

from servicelayer.util import pack_now, unpack_datetime
from servicelayer.util import dump_json, load_json
from servicelayer.cache import make_key

from memorious.core import conn
from memorious.model.crawl import Crawl

log = logging.getLogger(__name__)


class Event(object):
    """Document errors and warnings caused during tasks."""
    LEVEL_WARNING = 'warning'
    LEVEL_ERROR = 'error'
    LEVELS = [LEVEL_WARNING, LEVEL_ERROR]

    @classmethod
    def save(cls, crawler, stage, level, run_id, error=None, message=None):
        """Create an event, possibly based on an exception."""
        event = {
            'stage': stage.name,
            'level': level,
            'timestamp': pack_now(),
            'error': error,
            'message': message
        }
        data = dump_json(event)
        conn.lpush(make_key(crawler, "events"), data)
        conn.lpush(make_key(crawler, "events", level), data)
        conn.lpush(make_key(crawler, "events", stage), data)
        conn.lpush(make_key(crawler, "events", stage, level), data)
        conn.lpush(make_key(crawler, "events", run_id), data)
        conn.lpush(make_key(crawler, "events", run_id, level), data)
        return event

    @classmethod
    def delete(cls, crawler):
        conn.delete(make_key(crawler, "events"))
        for level in cls.LEVELS:
            conn.delete(make_key(crawler, "events", level))
        for run_id in Crawl.run_ids(crawler):
            conn.delete(make_key(crawler, "events", run_id))
            for level in cls.LEVELS:
                conn.delete(make_key(crawler, "events", run_id, level))
        for stage in crawler.stages.keys():
            conn.delete(make_key(crawler, "events", stage))
            for level in cls.LEVELS:
                conn.delete(make_key(crawler, "events", stage, level))

    @classmethod
    def get_counts(cls, crawler):
        counts = {}
        for level in cls.LEVELS:
            key = make_key(crawler, "events", level)
            counts[level] = conn.llen(key) or 0
        return counts

    @classmethod
    def get_stage_counts(cls, crawler, stage):
        counts = {}
        for level in cls.LEVELS:
            key = make_key(crawler, "events", stage, level)
            counts[level] = conn.llen(key) or 0
        return counts

    @classmethod
    def get_run_counts(cls, crawler, run_id):
        counts = {}
        for level in cls.LEVELS:
            key = make_key(crawler, "events", run_id, level)
            counts[level] = conn.llen(key) or 0
        return counts

    @classmethod
    def event_list(cls, key, start, end):
        results = []
        events = conn.lrange(key, start, end)
        if events is None:
            return results
        for event in events:
            result = load_json(event)
            result["timestamp"] = unpack_datetime(result['timestamp'])
            results.append(result)
        return results

    @classmethod
    def get_crawler_events(cls, crawler, start, end, level=None):
        key = make_key(crawler, "events", level)
        return cls.event_list(key, start, end)

    @classmethod
    def get_stage_events(cls, crawler, stage_name, start, end, level=None):
        """events from a particular stage"""
        key = make_key(crawler, "events", stage_name, level)
        return cls.event_list(key, start, end)

    @classmethod
    def get_run_events(cls, crawler, run_id, start, end, level=None):
        """Events from a particular run"""
        key = make_key(crawler, "events", run_id, level)
        return cls.event_list(key, start, end)



from banal import ensure_list
from lxml import html


def _get_html_document(context, data):
    with context.http.rehash(data) as result:
        if result.ok:
            return result.html


def clean_html(context, data):
    """Clean an HTML DOM and store the changed version."""
    doc = _get_html_document(context, data)
    if doc is None:
        context.emit(data=data)
        return

    remove_paths = context.params.get('remove_paths')
    for path in ensure_list(remove_paths):
        for el in doc.findall(path):
            el.drop_tree()

    html_text = html.tostring(doc, pretty_print=True)
    content_hash = context.store_data(html_text)
    data['content_hash'] = content_hash
    context.emit(data=data)



import logging
import dataset
from servicelayer import settings as sls
from servicelayer.archive import init_archive
from servicelayer.rate_limit import RateLimit
from servicelayer.cache import get_redis, get_fakeredis
from servicelayer.extensions import get_extensions
from sqlalchemy.pool import NullPool
from werkzeug.local import LocalProxy

from memorious import settings

log = logging.getLogger(__name__)


def load_manager():
    if not hasattr(settings, '_manager'):
        from memorious.logic.manager import CrawlerManager
        settings._manager = CrawlerManager()
        if settings.CONFIG_PATH:
            settings._manager.load_path(settings.CONFIG_PATH)
    return settings._manager


def load_datastore():
    if not hasattr(settings, '_datastore'):
        if not settings.DATASTORE_URI:
            raise RuntimeError("No $MEMORIOUS_DATASTORE_URI.")
        # do not pool connections for the datastore
        engine_kwargs = {'poolclass': NullPool}
        settings._datastore = dataset.connect(settings.DATASTORE_URI,
                                              engine_kwargs=engine_kwargs)
        # Use bigint to store integers by default
        settings._datastore.types.integer = settings._datastore.types.bigint
    return settings._datastore


def is_sync_mode():
    if settings.TESTING or settings.DEBUG:
        return True
    return sls.REDIS_URL is None


def connect_redis():
    if settings.TESTING:
        return get_fakeredis()
    return get_redis()


manager = LocalProxy(load_manager)
datastore = LocalProxy(load_datastore)
conn = LocalProxy(connect_redis)

# File storage layer for blobs on local file system or S3
storage = init_archive()


def init_memorious():
    for func in get_extensions('memorious.plugins'):
        func()


def get_rate_limit(resource, limit=100, interval=60, unit=1):
    return RateLimit(conn, resource, limit=limit, interval=interval, unit=unit)



import os
import json

import pytest

from memorious.operations.fetch import fetch, session
from memorious.operations.parse import parse
from memorious.operations.initializers import seed, sequence, dates, enumerate
from memorious.operations.store import directory
from memorious.core import connect_redis


@pytest.mark.parametrize("url,call_count", [
    ("https://httpbin.org/html", 1),
    ("https://occrp.org/", 0),
    ("https://httpbin.org/status/418", 0)
])
def test_fetch(url, call_count, context, mocker):
    rules = {'pattern': 'https://httpbin.org/*'}
    context.params["rules"] = rules
    mocker.patch.object(context, "emit")
    fetch(context, {"url": url})
    assert context.emit.call_count == call_count


def test_session(context, mocker):
    context.params["user"] = "user"
    context.params["password"] = "password"
    context.params["user_agent"] = "Godzilla Firehose 0.1"
    context.params["url"] = "https://httpbin.org/get"
    data = {"hello": "world"}
    mocker.patch.object(context.http, "save")
    mocker.patch.object(context, "emit")

    session(context, data)

    assert context.http.save.called_one_with()
    assert context.emit.called_one_with(data=data)
    assert (
        context.http.session.headers["User-Agent"] == "Godzilla Firehose 0.1"
    )
    assert context.http.session.headers["Referer"] == "https://httpbin.org/get"
    assert context.http.session.auth == ("user", "password")


def test_parse(context, mocker):
    url = "http://example.org/"
    result = context.http.get(url)
    data = result.serialize()

    mocker.patch.object(context, "emit")

    rules = {'pattern': 'https://httpbin.org/*'}
    context.params["store"] = rules
    parse(context, data)
    assert context.emit.call_count == 1
    context.emit.assert_called_once_with(rule="fetch", data={
        "url": "https://www.iana.org/domains/example"
    })

    # cleanup tags
    conn = connect_redis()
    conn.flushall()

    context.http.result = None
    context.params["store"] = None
    parse(context, data)
    assert context.emit.call_count == 3, data


def test_seed(context, mocker):
    context.params["url"] = None
    context.params["urls"] = [
        "http://httpbin.org/status/%(status)s"
    ]
    mocker.patch.object(context, "emit")
    seed(context, data={"status": 404})
    assert context.emit.call_count == 1
    context.emit.assert_called_once_with(
        data={"url": "http://httpbin.org/status/404"}
    )


def test_sequence(context, mocker):
    mocker.patch.object(context, "emit")

    context.params["start"] = 2
    context.params["stop"] = 11
    context.params["step"] = 3
    sequence(context, data={})
    assert context.emit.call_count == 3

    context.params["start"] = 7
    context.params["stop"] = 1
    context.params["step"] = -3
    sequence(context, data={})
    assert context.emit.call_count == 5


def test_dates(context, mocker):
    mocker.patch.object(context, "emit")
    mocker.patch.object(context, "recurse")
    context.params["format"] = '%d-%m-%Y'
    context.params["days"] = 3
    context.params["begin"] = "10-12-2012"
    context.params["end"] = "20-12-2012"
    dates(context, data={})
    assert context.emit.call_count == 1
    context.emit.assert_called_once_with(data={
        "date": "20-12-2012",
        "date_iso": "2012-12-20T00:00:00"
    })
    assert context.recurse.call_count == 1
    context.recurse.assert_called_once_with(
        data={"current": "17-12-2012"}
    )


def test_enumerate(context, mocker):
    mocker.patch.object(context, "emit")
    context.params["items"] = [1, 2, 3]
    enumerate(context, data={})
    assert context.emit.call_count == 3
    # expected_calls = [
    #     mocker.call(data={"item": 1}),
    #     mocker.call(data={"item": 2}),
    #     mocker.call(data={"item": 3}),
    # ]
    # assert context.emit.mock_calls == expected_calls
    # Ideally this should work, but currently it doesn't; because we pass
    # the reference to data dict around and then mutate it.


def test_directory(context):
    # echo user-agent
    url = "https://httpbin.org/user-agent"
    result = context.http.get(url, headers={"User-Agent": "Memorious Test"})
    data = result.serialize()
    directory(context, data)

    content_hash = data.get('content_hash')
    file_path = os.path.realpath(__file__)
    store_dir = os.path.normpath(os.path.join(
        file_path, "../testdata/data/store/occrp_web_site"
    ))

    raw_file_path = os.path.join(store_dir, content_hash + ".data.json")
    meta_file_path = os.path.join(store_dir, content_hash + ".json")
    assert os.path.exists(meta_file_path)
    assert os.path.exists(raw_file_path)

    with open(meta_file_path, "rb") as fh:
        assert json.load(fh)['content_hash'] == data['content_hash']
    with open(raw_file_path, "rb") as fh:
        assert b'"user-agent": "Memorious Test"' in fh.read()



import logging
import requests
import urllib3
from requests.packages.urllib3.exceptions import InsecureRequestWarning

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
logging.getLogger('requests').setLevel(logging.WARNING)
logging.getLogger('zeep').setLevel(logging.WARNING)
logging.getLogger('httpstream').setLevel(logging.WARNING)
logging.getLogger('urllib3').setLevel(logging.WARNING)
logging.getLogger('rdflib').setLevel(logging.WARNING)
logging.getLogger('chardet').setLevel(logging.WARNING)



from importlib import import_module
from servicelayer.extensions import get_entry_point

from memorious.model import Crawl


class CrawlerStage(object):
    """A single step in a data processing crawler."""

    def __init__(self, crawler, name, config):
        self.crawler = crawler
        self.name = name
        self.config = config
        self.method_name = config.get('method')
        self.params = config.get('params') or {}
        self.handlers = config.get('handle') or {}

    @property
    def method(self):
        # method A: via a named Python entry point
        func = get_entry_point('memorious.operations', self.method_name)
        if func is not None:
            return func
        # method B: direct import from a module
        if ':' not in self.method_name:
            raise ValueError("Unknown method: %s", self.method_name)
        package, method = self.method_name.rsplit(':', 1)
        module = import_module(package)
        return getattr(module, method)

    @property
    def op_count(self):
        """Total operations performed for this stage"""
        return Crawl.op_count(self.crawler, self)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<CrawlerStage(%r, %s)>' % (self.crawler, self.name)



import os
import io
import yaml
import logging
from datetime import timedelta, datetime
from importlib import import_module
from servicelayer.jobs import Dataset, Job

from memorious import settings
from memorious.core import conn
from memorious.model import Event, Crawl, Queue
from memorious.logic.stage import CrawlerStage

log = logging.getLogger(__name__)


class Crawler(object):
    """A processing graph that constitutes a crawler."""
    SCHEDULES = {
        'hourly': timedelta(hours=1),
        'daily': timedelta(days=1),
        'weekly': timedelta(weeks=1),
        'monthly': timedelta(weeks=4)
    }

    def __init__(self, manager, source_file):
        self.manager = manager
        self.source_file = source_file
        with io.open(source_file, encoding='utf-8') as fh:
            self.config_yaml = fh.read()
            self.config = yaml.safe_load(self.config_yaml)

        self.name = os.path.basename(source_file)
        self.name = self.config.get('name', self.name)
        self.description = self.config.get('description', self.name)
        self.category = self.config.get('category', 'scrape')
        self.schedule = self.config.get('schedule')
        self.disabled = self.config.get('disabled', False)
        self.init_stage = self.config.get('init', 'init')
        self.delta = Crawler.SCHEDULES.get(self.schedule)
        self.delay = int(self.config.get('delay', 0))
        self.expire = int(self.config.get('expire', settings.EXPIRE)) * 84600
        self.stealthy = self.config.get('stealthy', False)
        self.queue = Dataset(conn, self.name)
        self.aggregator_config = self.config.get('aggregator', {})

        self.stages = {}
        for name, stage in self.config.get('pipeline', {}).items():
            self.stages[name] = CrawlerStage(self, name, stage)

    def check_due(self):
        """Check if the last execution of this crawler is older than
        the scheduled interval."""
        if self.disabled:
            return False
        if self.is_running:
            return False
        if self.delta is None:
            return False
        last_run = self.last_run
        if last_run is None:
            return True
        now = datetime.utcnow()
        if now > last_run + self.delta:
            return True
        return False

    @property
    def aggregator_method(self):
        if self.aggregator_config:
            method = self.aggregator_config.get("method")
            if not method:
                return
            if ':' in method:
                package, method = method.rsplit(':', 1)
                module = import_module(package)
                return getattr(module, method)

    def aggregate(self, context):
        if self.aggregator_method:
            log.info("Running aggregator for %s" % self.name)
            params = self.aggregator_config.get("params", {})
            self.aggregator_method(context, params)

    def flush(self):
        """Delete all run-time data generated by this crawler."""
        self.queue.cancel()
        Event.delete(self)
        Crawl.flush(self)

    def flush_events(self):
        Event.delete(self)

    def cancel(self):
        Crawl.abort_all(self)
        self.queue.cancel()

    def run(self, incremental=None, run_id=None):
        """Queue the execution of a particular crawler."""
        state = {
            'crawler': self.name,
            'run_id': run_id or Job.random_id(),
            'incremental': settings.INCREMENTAL
        }
        if incremental is not None:
            state['incremental'] = incremental

        # Cancel previous runs:
        self.cancel()
        # Flush out previous events:
        Event.delete(self)
        Queue.queue(self.init_stage, state, {})

    @property
    def is_running(self):
        """Is the crawler currently running?"""
        if self.disabled:
            return False
        for job in self.queue.get_jobs():
            if not job.is_done():
                return True
        return False

    @property
    def last_run(self):
        return Crawl.last_run(self)

    @property
    def op_count(self):
        """Total operations performed for this crawler"""
        return Crawl.op_count(self)

    @property
    def runs(self):
        return Crawl.runs(self)

    @property
    def latest_runid(self):
        return Crawl.latest_runid(self)

    @property
    def pending(self):
        status = self.queue.get_status()
        return status.get('pending')

    def get(self, name):
        return self.stages.get(name)

    def __str__(self):
        return self.name

    def __iter__(self):
        return iter(self.stages.values())

    def __repr__(self):
        return '<Crawler(%s)>' % self.name





class ViewForm(dict):
    """Helper for VIEWSTATE in ASP-driven web sites."""

    def __init__(self, element=None, data=None):
        if data is not None:
            self.update(data)
        self.element = element
        for inp in element.findall('.//input'):
            name = inp.get('name')
            if name is None:
                continue
            self[name] = inp.get('value', '')

    def clear(self, name):
        self.pop(name, None)

    @classmethod
    def from_result(cls, result):
        return cls(element=result.html)



import os
import pkg_resources
import multiprocessing
from servicelayer import env
from servicelayer import settings as sls

###############################################################################
# Core configuration
VERSION = pkg_resources.get_distribution('memorious').version
APP_NAME = env.get('MEMORIOUS_APP_NAME', 'memorious')

# Enable debug logging etc.
DEBUG = env.to_bool('MEMORIOUS_DEBUG', default=False)
TESTING = False

# Base operating path
BASE_PATH = os.path.join(os.getcwd(), 'data')
BASE_PATH = env.get('MEMORIOUS_BASE_PATH', BASE_PATH)

# Override servicelayer archive if undefined
sls.ARCHIVE_PATH = sls.ARCHIVE_PATH or os.path.join(BASE_PATH, 'archive')

# Directory which contains crawler pipeline YAML specs
CONFIG_PATH = env.get('MEMORIOUS_CONFIG_PATH')

# Try and run scrapers in a way that only acquires new data
INCREMENTAL = env.to_bool('MEMORIOUS_INCREMENTAL', default=True)

# How many days until an incremental crawl expires
EXPIRE = env.to_int('MEMORIOUS_EXPIRE', 60)

# How many db inserts per minute
DB_RATE_LIMIT = env.to_int('MEMORIOUS_DB_RATE_LIMIT', 6000)

# How many http requests to a host per minute
HTTP_RATE_LIMIT = env.to_int('MEMORIOUS_HTTP_RATE_LIMIT', 120)  # noqa

# How many seconds to wait before trying to run scheduled crawlers
SCHEDULER_INTERVAL = env.to_int('MEMORIOUS_SCHEDULER_INTERVAL', 60)

# How many threads to use for execution
THREADS = env.to_int('MEMORIOUS_THREADS', min(8, multiprocessing.cpu_count()))

# Max scheduled tasks at the same time
MAX_SCHEDULED = env.to_int('MEMORIOUS_MAX_SCHEDULED', THREADS)

# Max number of tasks in a stage's task queue
MAX_QUEUE_LENGTH = env.to_int('MEMORIOUS_MAX_QUEUE_LENGTH', 50000)

# HTTP request configuration
HTTP_CACHE = env.to_bool('MEMORIOUS_HTTP_CACHE', default=True)

# HTTP user agent default
USER_AGENT = 'Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US; rv:1.1)'
USER_AGENT = '%s aleph.memorious/%s' % (USER_AGENT, VERSION)
USER_AGENT = env.get('MEMORIOUS_USER_AGENT', USER_AGENT)

# Datastore: operational data store (ODS) database connection
DATASTORE_FILE = os.path.join(BASE_PATH, 'datastore.sqlite3')
DATASTORE_URI = 'sqlite:///%s' % DATASTORE_FILE
DATASTORE_URI = env.get('MEMORIOUS_DATASTORE_URI', DATASTORE_URI)



import os
from uuid import uuid4


def random_filename(path=None):
    """Make a UUID-based file name which is extremely unlikely
    to exist already."""
    filename = uuid4().hex
    if path is not None:
        filename = os.path.join(path, filename)
    return filename



import datetime

from memorious.core import conn
from memorious.model import Crawl


class TestReporting(object):

    def test_operation_reporting(self, crawler, context):
        conn.flushall()
        assert len(crawler.runs) == 0
        stage = list(crawler.stages)[0]
        Crawl.operation_start(crawler, stage, context.run_id)
        assert crawler.latest_runid == context.run_id
        assert len(crawler.runs) == 1
        assert crawler.op_count == 1
        assert isinstance(crawler.last_run, datetime.datetime)
        Crawl.operation_end(crawler, context.run_id)
        crawler.flush()
        assert crawler.op_count == 0
        assert len(crawler.runs) == 0



class MemoriousException(Exception):
    """Base exception class."""
    pass


class QueueTooBigError(MemoriousException):
    pass


class ConfigurationError(MemoriousException):
    """A configuration option is not set."""


class RuleParsingException(MemoriousException):
    """A rule encounters something it can't parse."""
    pass


class StorageFileMissing(MemoriousException):
    """A file could not be found in the blob storage."""

    def __init__(self, content_hash, file_name=None):
        self.content_hash = content_hash
        self.file_name = file_name
        msg = 'Could not load: %s' % content_hash
        super(StorageFileMissing, self).__init__(msg)


class ParseError(MemoriousException):
    """An error while parsing a structured HTTP response."""
    pass


class RateLimitException(Exception):
    """Rate limit exceeded for a crawler operation"""
    pass



import os
from datetime import datetime
from urllib.parse import urljoin

DEFAULT_HOST = 'https://documentcloud.org/'
DEFAULT_INSTANCE = 'documentcloud'

LANGUAGES = {
    "ara": "ar",
    "zho": "zh",
    "tra": "zh",
    "hrv": "hr",
    "dan": "da",
    "nld": "nl",
    "eng": "en",
    "fra": "fr",
    "deu": "de",
    "heb": "he",
    "hun": "hu",
    "ind": "id",
    "ita": "it",
    "jpn": "ja",
    "kor": "ko",
    "nor": "no",
    "por": "pt",
    "ron": "ro",
    "rus": "ru",
    "spa": "es",
    "swe": "sv",
    "ukr": "uk"
}


def documentcloud_query(context, data):
    host = context.get('host', DEFAULT_HOST)
    instance = context.get('instance', DEFAULT_INSTANCE)
    query = context.get('query')
    page = data.get('page', 1)

    search_url = urljoin(host, 'search/documents.json')
    res = context.http.get(search_url, params={
        'q': query,
        'per_page': 100,
        'page': page
    })
    documents = res.json.get('documents', [])
    for document in documents:
        doc = {
            'foreign_id': '%s:%s' % (instance, document.get('id')),
            'url': document.get('pdf_url'),
            'source_url': document.get('canonical_url'),
            'title': document.get('title'),
            'author': document.get('author'),
            'file_name': os.path.basename(document.get('pdf_url')),
            'mime_type': 'application/pdf'
        }

        lang = LANGUAGES.get(document.get('language'))
        if lang is not None:
            data['languages'] = [lang]

        published = document.get('created_at')
        if published is not None:
            dt = datetime.strptime(published, '%b %d, %Y')
            doc['published_at'] = dt.isoformat()

        context.emit(data=doc)

    if len(documents):
        context.recurse(data={'page': page + 1})



import logging
from banal import ensure_list
from datetime import datetime

from servicelayer.jobs import Job
from servicelayer.util import unpack_int, unpack_datetime, pack_now
from servicelayer.cache import make_key

from memorious.core import conn

log = logging.getLogger(__name__)


class Crawl(object):
    """The current state of a running crawler instance"""

    @classmethod
    def last_run(cls, crawler):
        last_run = conn.get(make_key(crawler, "last_run"))
        return unpack_datetime(last_run)

    @classmethod
    def op_count(cls, crawler, stage=None):
        """Total operations performed for this crawler"""
        if stage:
            total_ops = conn.get(make_key(crawler, stage))
        else:
            total_ops = conn.get(make_key(crawler, "total_ops"))
        return unpack_int(total_ops)

    @classmethod
    def run_ids(cls, crawler):
        key = make_key(crawler, "runs")
        return ensure_list(conn.smembers(key))

    @classmethod
    def runs(cls, crawler):
        runs = []
        for run_id in cls.run_ids(crawler):
            start = conn.get(make_key("run", run_id, "start"))
            end = conn.get(make_key("run", run_id, "end"))
            total_ops = conn.get(make_key("run", run_id, "total_ops"))
            runs.append({
                'run_id': run_id,
                'total_ops': unpack_int(total_ops),
                'start': unpack_datetime(start, datetime.utcnow()),
                'end': unpack_datetime(end)
            })
        return runs

    @classmethod
    def operation_start(cls, crawler, stage, run_id):
        if not conn.sismember(make_key(crawler, "runs"), run_id):
            conn.sadd(make_key(crawler, "runs"), run_id)
            conn.set(make_key("run", run_id, "start"), pack_now())
        conn.incr(make_key("run", run_id))
        conn.incr(make_key("run", run_id, "total_ops"))
        conn.incr(make_key(crawler, stage))
        conn.incr(make_key(crawler, "total_ops"))
        conn.set(make_key(crawler, "last_run"), pack_now())
        conn.set(make_key(crawler, "current_run"), run_id)

    @classmethod
    def operation_end(cls, crawler, run_id):
        conn.set(make_key(crawler, "last_run"), pack_now())
        pending = conn.decr(make_key("run", run_id))
        if unpack_int(pending) == 0:
            conn.set(make_key("run", run_id, "end"), pack_now())

    @classmethod
    def flush(cls, crawler):
        for stage in crawler.stages:
            conn.delete(make_key(crawler, stage))

        for run_id in cls.run_ids(crawler):
            conn.delete(make_key(crawler, run_id))
            conn.delete(make_key(crawler, run_id, "start"))
            conn.delete(make_key(crawler, run_id, "end"))
            conn.delete(make_key(crawler, run_id, "total_ops"))

        conn.delete(make_key(crawler, "runs"))
        conn.delete(make_key(crawler, "current_run"))
        conn.delete(make_key(crawler, "total_ops"))
        conn.delete(make_key(crawler, "last_run"))
        conn.delete(make_key(crawler, "runs_abort"))

    @classmethod
    def latest_runid(cls, crawler):
        return conn.get(make_key(crawler, "current_run"))

    @classmethod
    def abort_run(cls, crawler, run_id):
        conn.sadd(make_key(crawler, "runs_abort"), run_id)
        conn.setnx(make_key("run", run_id, "end"), pack_now())
        job = Job(conn, crawler.queue, run_id)
        job.remove()

    @classmethod
    def abort_all(cls, crawler):
        for run_id in cls.run_ids(crawler):
            cls.abort_run(crawler, run_id)

    @classmethod
    def is_aborted(cls, crawler, run_id):
        key = make_key(crawler, "runs_abort")
        return conn.sismember(key, run_id)



import cgi
import json
import pickle
import codecs
from hashlib import sha1
from pathlib import Path
from lxml import html, etree
from datetime import datetime, timedelta
from urllib.parse import unquote, urlparse
from banal import hash_data, is_mapping
from urlnormalizer import normalize_url
from pantomime import parse_mimetype, normalize_mimetype
from normality import guess_file_encoding, stringify
from requests import Session, Request
from requests.structures import CaseInsensitiveDict
from servicelayer.cache import make_key
from servicelayer.util import QUEUE_EXPIRE

from memorious import settings
from memorious.core import conn, storage, get_rate_limit
from memorious.logic.mime import NON_HTML
from memorious.exc import ParseError
from memorious.helpers.ua import UserAgent
from memorious.helpers.dates import parse_date
from memorious.util import random_filename


class ContextHttp(object):
    STATE_SESSION = '_http'

    def __init__(self, context):
        self.context = context

        self.cache = settings.HTTP_CACHE
        if 'cache' in context.params:
            self.cache = context.params.get('cache')

        self.session = self.load_session()
        if self.session is None:
            self.reset()

    def reset(self):
        self.session = Session()
        self.session.headers['User-Agent'] = settings.USER_AGENT
        if self.context.crawler.stealthy:
            self.session.headers['User-Agent'] = UserAgent().random()
        return self.session

    def request(self, method, url, headers={}, auth=None, data=None,
                params=None, json=None, allow_redirects=True, lazy=False):
        if is_mapping(params):
            params = list(params.items())

        method = method.upper().strip()
        request = Request(method, url, data=data, headers=headers,
                          json=json, auth=auth)
        response = ContextHttpResponse(self,
                                       request=request,
                                       allow_redirects=allow_redirects)
        if not lazy:
            response.fetch()
        return response

    def get(self, url, **kwargs):
        return self.request('GET', url, **kwargs)

    def post(self, url, **kwargs):
        return self.request('POST', url, **kwargs)

    def rehash(self, data):
        return ContextHttpResponse.deserialize(self, data)

    def load_session(self):
        if self.STATE_SESSION not in self.context.state:
            return
        key = self.context.state.get(self.STATE_SESSION)
        value = conn.get(key)
        if value is not None:
            session = codecs.decode(bytes(value, 'utf-8'), 'base64')
            return pickle.loads(session)

    def save(self):
        session = pickle.dumps(self.session)
        session = codecs.encode(session, 'base64')
        key = sha1(session).hexdigest()[:15]
        key = make_key(self.context.run_id, "session", key)
        conn.set(key, session, ex=QUEUE_EXPIRE)
        self.context.state[self.STATE_SESSION] = key


class ContextHttpResponse(object):
    """Handle a cached and managed HTTP response.

    This is a wrapper for ``requests`` HTTP response which adds several
    aspects:

    * Uses HTTP caching against the database when configured to do so.
    * Will evaluate lazily in order to allow fast web crawling.
    * Allow responses to be serialized between crawler operations.
    """
    CACHE_METHODS = ['GET', 'HEAD']

    def __init__(self, http, request=None, allow_redirects=True):
        self.http = http
        self.context = http.context
        self.request = request
        self.allow_redirects = allow_redirects
        self._response = None
        self._status_code = None
        self._url = None
        self._request_id = None
        self._headers = None
        self._encoding = None
        self._content_hash = None
        self._file_path = None
        self.retrieved_at = None
        self._remove_file = False

    @property
    def use_cache(self):
        # It's complicated.
        if not self.http.cache:
            return False
        if self.request is not None:
            return self.request.method in self.CACHE_METHODS
        return True

    @property
    def response(self):
        if self._response is None and self.request is not None:
            request = self.request
            existing = None
            if self.use_cache:
                existing = self.context.get_tag(self.request_id)
            if existing is not None:
                headers = CaseInsensitiveDict(existing.get('headers'))
                last_modified = headers.get('last-modified')
                if last_modified:
                    request.headers['If-Modified-Since'] = last_modified

                etag = headers.get('etag')
                if etag:
                    request.headers['If-None-Match'] = etag

            self._rate_limit(request.url)

            session = self.http.session
            prepared = session.prepare_request(request)
            response = session.send(prepared,
                                    stream=True,
                                    verify=False,
                                    allow_redirects=self.allow_redirects)

            if existing is not None and response.status_code == 304:
                self.context.log.info("Using cached HTTP response: %s",
                                      response.url)
                self.apply_data(existing)
            else:
                self._response = response

            # update the serialised session with cookies etc.
            self.http.save()
        return self._response

    def fetch(self):
        """Lazily trigger download of the data when requested."""
        if self._file_path is not None:
            return self._file_path
        temp_path = self.context.work_path
        if self._content_hash is not None:
            self._file_path = storage.load_file(self._content_hash,
                                                temp_path=temp_path)
            return self._file_path
        if self.response is not None:
            self._file_path = random_filename(temp_path)
            content_hash = sha1()
            with open(self._file_path, 'wb') as fh:
                for chunk in self.response.iter_content(chunk_size=8192):
                    content_hash.update(chunk)
                    fh.write(chunk)
            self._remove_file = True
            chash = content_hash.hexdigest()
            self._content_hash = storage.archive_file(self._file_path,
                                                      content_hash=chash)
            if self.http.cache and self.ok:
                self.context.set_tag(self.request_id, self.serialize())
            self.retrieved_at = datetime.utcnow().isoformat()
        return self._file_path

    def _complete(self):
        if self._content_hash is None:
            self.fetch()

    def _rate_limit(self, url):
        resource = urlparse(url).netloc or url
        limit = self.context.get('http_rate_limit', settings.HTTP_RATE_LIMIT)
        rate_limit = get_rate_limit(resource, limit=limit)
        rate_limit.comply()

    @property
    def url(self):
        if self._url is not None:
            return self._url
        if self._response is not None:
            return normalize_url(self._response.url)
        if self.request is not None:
            return self.request.url

    @property
    def request_id(self):
        if self._request_id is not None:
            return self._request_id
        if self.request is not None:
            parts = [self.request.method, normalize_url(self.url)]
            if self.request.data:
                parts.append(hash_data(self.request.data))
            if self.request.json:
                parts.append(hash_data(self.request.json))
            return make_key(*parts)

    @property
    def status_code(self):
        if self._status_code is None and self.response is not None:
            self._status_code = self.response.status_code
        return self._status_code

    @property
    def headers(self):
        if self._headers is None and self.response:
            self._headers = self.response.headers
        return self._headers or CaseInsensitiveDict()

    @property
    def last_modified(self):
        now = datetime.utcnow()
        last_modified_header = self.headers.get("Last-Modified")
        if last_modified_header is not None:
            # Tue, 15 Nov 1994 12:45:26 GMT
            last_modified = parse_date(last_modified_header)
            if last_modified < now + timedelta(seconds=30):
                return last_modified.strftime("%Y-%m-%dT%H:%M:%S%z")
        return None

    @property
    def encoding(self):
        if self._encoding is None:
            mime = parse_mimetype(self.headers.get('content-type'))
            self._encoding = mime.charset
        if self._encoding is None:
            with open(self.file_path, 'rb') as fh:
                self._encoding = guess_file_encoding(fh)
        return self._encoding

    @encoding.setter
    def encoding(self, encoding):
        self._encoding = encoding

    @property
    def file_path(self):
        self.fetch()
        if self._file_path is not None:
            return Path(self._file_path)

    @property
    def content_hash(self):
        self.fetch()
        return self._content_hash

    @property
    def content_type(self):
        content_type = self.headers.get('content-type')
        return normalize_mimetype(content_type)

    @property
    def file_name(self):
        disposition = self.headers.get('content-disposition')
        file_name = None
        if disposition is not None:
            _, options = cgi.parse_header(disposition)
            filename = options.get('filename') or ''
            file_name = stringify(unquote(filename))
        return file_name

    @property
    def ok(self):
        if self.status_code is None:
            return False
        return self.status_code < 400

    @property
    def raw(self):
        if not hasattr(self, '_raw'):
            self._raw = None
            if self.file_path is not None:
                with open(self.file_path, 'rb') as fh:
                    self._raw = fh.read()
        return self._raw

    @property
    def text(self):
        if self.raw is None:
            return None
        return self.raw.decode(self.encoding, 'replace')

    @property
    def html(self):
        if not hasattr(self, '_html'):
            self._html = None
            if self.content_type in NON_HTML:
                return
            if self.raw is None or not len(self.raw):
                return
            try:
                self._html = html.fromstring(self.text)
            except ValueError as ve:
                if 'encoding declaration' in str(ve):
                    self._html = html.parse(self.file_path.as_posix())
            except (etree.ParserError, etree.ParseError):
                pass
        return self._html

    @property
    def xml(self):
        if not hasattr(self, '_xml'):
            parser = etree.XMLParser(
                ns_clean=True,
                recover=True,
                resolve_entities=False,
                no_network=True
            )
            self._xml = etree.parse(self.file_path.as_posix(), parser=parser)
        return self._xml

    @property
    def json(self):
        if not hasattr(self, '_json'):
            if self.file_path is None:
                raise ParseError("Cannot parse failed download.")
            with open(self.file_path, 'r') as fh:
                self._json = json.load(fh)
        return self._json

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        if self._response is not None:
            self._response.close()
        # Will be deleted by the context:
        # storage.cleanup_file(self._content_hash)

    def serialize(self):
        self.fetch()
        data = {
            'request_id': self.request_id,
            'status_code': self.status_code,
            'url': self.url,
            'content_hash': self.content_hash,
            'encoding': self._encoding,
            'headers': dict(self.headers),
            'retrieved_at': self.retrieved_at
        }
        if self.last_modified is not None:
            data['modified_at'] = self.last_modified
        return data

    def apply_data(self, data):
        self._status_code = data.get('status_code')
        self._url = data.get('url')
        self._request_id = data.get('request_id')
        self._headers = CaseInsensitiveDict(data.get('headers'))
        self._encoding = data.get('encoding')
        self._content_hash = data.get('content_hash')
        self.retrieved_at = data.get('retrieved_at')

    @classmethod
    def deserialize(cls, http, data):
        obj = cls(http)
        obj.apply_data(data)
        return obj

    def __repr__(self):
        return '<ContextHttpResponse(%s,%s)>' % (self.url,
                                                 self._content_hash)



from banal import ensure_list
from datetime import datetime, timedelta


def seed(context, data):
    """Initialize a crawler with a set of seed URLs.

    The URLs are given as a list or single value to the ``urls`` parameter.

    If this is called as a second stage in a crawler, the URL will be formatted
    against the supplied ``data`` values, e.g.:

        https://crawl.site/entries/%(number)s.html
    """
    for key in ('url', 'urls'):
        for url in ensure_list(context.params.get(key)):
            url = url % data
            context.emit(data={'url': url})


def enumerate(context, data):
    """Iterate through a set of items and emit each one of them."""
    items = ensure_list(context.params.get('items'))
    for item in items:
        data['item'] = item
        context.emit(data=data)


def sequence(context, data):
    """Generate a sequence of numbers.

    It is the memorious equivalent of the xrange function, accepting the
    ``start``, ``stop`` and ``step`` parameters.

    This can run in two ways:
    * As a single function generating all numbers in the given range.
    * Recursively, generating numbers one by one with an optional ``delay``.

    The latter mode is useful in order to generate very large sequences
    without completely clogging up the user queue.

    If an optional ``tag`` is given, each number will be emitted only once
    across multiple runs of the crawler.
    """
    number = data.get('number', context.params.get('start', 1))
    stop = context.params.get('stop')
    step = context.params.get('step', 1)
    delay = context.params.get('delay')
    prefix = context.params.get('tag')
    while True:
        tag = None if prefix is None else '%s:%s' % (prefix, number)

        if tag is None or not context.check_tag(tag):
            context.emit(data={'number': number})

        if tag is not None:
            context.set_tag(tag, True)

        number = number + step
        if step > 0 and number >= stop:
            break
        if step < 0 and number <= stop:
            break

        if delay is not None:
            context.recurse(data={'number': number}, delay=delay)
            break


def dates(context, data):
    format = context.params.get('format', '%Y-%m-%d')
    delta = timedelta(days=context.params.get('days', 0),
                      weeks=context.params.get('weeks', 0))
    if delta == timedelta():
        context.log.error("No interval given: %r", context.params)
        return

    if 'end' in context.params:
        current = context.params.get('end')
        current = datetime.strptime(current, format)
    else:
        current = datetime.utcnow()

    if 'current' in data:
        current = datetime.strptime(data.get('current'), format)

    if 'begin' in context.params:
        begin = context.params.get('begin')
        begin = datetime.strptime(begin, format)
    else:
        steps = context.params.get('steps', 100)
        begin = current - (delta * steps)

    context.emit(data={
        'date': current.strftime(format),
        'date_iso': current.isoformat()
    })
    current = current - delta
    if current >= begin:
        data = {'current': current.strftime(format)}
        context.recurse(data=data)



import click
import logging
import time
from tabulate import tabulate

from memorious import settings
from memorious.core import manager, init_memorious, is_sync_mode
from memorious.worker import get_worker

log = logging.getLogger(__name__)


@click.group()
@click.option('--debug/--no-debug', default=False,
              envvar='MEMORIOUS_DEBUG')
@click.option('--cache/--no-cache', default=True,
              envvar='MEMORIOUS_HTTP_CACHE')
@click.option('--incremental/--non-incremental', default=True,
              envvar='MEMORIOUS_INCREMENTAL')
def cli(debug, cache, incremental):
    """Crawler framework for documents and structured scrapers."""
    settings.HTTP_CACHE = cache
    settings.INCREMENTAL = incremental
    settings.DEBUG = debug
    if settings.DEBUG:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
    init_memorious()


def get_crawler(name):
    crawler = manager.get(name)
    if crawler is None:
        msg = 'Crawler [%s] not found.' % name
        raise click.BadParameter(msg, param=crawler)
    return crawler


@cli.command()
@click.argument('crawler')
def run(crawler):
    """Run a specified crawler."""
    crawler = get_crawler(crawler)
    crawler.run()
    if is_sync_mode():
        worker = get_worker()
        worker.sync()

@cli.command()
@click.argument('crawler')
def cancel(crawler):
    """Abort execution of a specified crawler."""
    crawler = get_crawler(crawler)
    crawler.cancel()


@cli.command()
@click.argument('crawler')
def flush(crawler):
    """Delete all data generated by a crawler."""
    crawler = get_crawler(crawler)
    crawler.flush()


@cli.command()
def process():
    """Start the queue and process tasks as they come. Blocks while waiting"""
    worker = get_worker()
    worker.run()


@cli.command('list')
def index():
    """List the available crawlers."""
    crawler_list = []
    for crawler in manager:
        is_due = 'yes' if crawler.check_due() else 'no'
        if crawler.disabled:
            is_due = 'off'
        crawler_list.append([crawler.name,
                             crawler.description,
                             crawler.schedule,
                             is_due,
                             crawler.pending])
    headers = ['Name', 'Description', 'Schedule', 'Due', 'Pending']
    print(tabulate(crawler_list, headers=headers))


@cli.command()
def killthekitten():
    """Completely kill redis contents."""
    from memorious.core import connect_redis
    conn = connect_redis()
    conn.flushall()


def main():
    cli(obj={})


if __name__ == '__main__':
    main()



import os
import pytest
from requests import Session, Response, Request
from lxml import html, etree

from memorious.logic.http import ContextHttpResponse


class TestContextHttp(object):
    def test_session(self, http):
        assert isinstance(http.session, Session)

    def test_response(self, http):
        response = http.get("https://httpbin.org/get")
        assert isinstance(response, ContextHttpResponse)
        assert isinstance(response._response, Response)

    def test_response_lazy(self, http):
        response = http.get("https://httpbin.org/get", lazy=True)
        assert isinstance(response, ContextHttpResponse)
        assert response._response is None


class TestContextHttpResponse(object):
    def test_fetch_response(self, http):
        request = Request("GET", "https://httpbin.org/get")
        context_http_response = ContextHttpResponse(http, request)
        file_path = context_http_response.fetch()
        assert os.path.exists(file_path)

    def test_contenttype(self, http):
        request = Request("GET", "https://httpbin.org/get")
        context_http_response = ContextHttpResponse(http, request)
        assert context_http_response.content_type == "application/json"

    def test_attachment(self, http):
        request = Request(
            "GET", "https://httpbin.org/response-headers?Content-Type="
            "text/plain;%20charset=UTF-8&Content-Disposition=attachment;"
            "%20filename%3d%22test.json%22"
        )
        context_http_response = ContextHttpResponse(http, request)
        assert context_http_response.file_name == "test.json"

    @pytest.mark.parametrize("url,encoding", [
        ("https://httpbin.org/response-headers?charset=", "utf-8"),
        ("https://httpbin.org/response-headers?content-type=text"
         "/plain;%20charset=utf-16", "utf-16"),
        ("https://httpbin.org/response-headers?Content-Type=text/"
         "plain;%20charset=utf-32", "utf-32")
    ])
    def test_encoding(self, url, encoding, http):
        request = Request("GET", url)
        context_http_response = ContextHttpResponse(http, request)
        assert context_http_response.encoding == encoding

    def test_request_id(self, http):
        request = Request(
            "GET", "https://httpbin.org/get", data={"hello": "world"}
        )
        context_http_response = ContextHttpResponse(http, request)
        assert context_http_response._request_id is None
        assert isinstance(context_http_response.request_id, str)

    def test_content(self, http):
        request = Request(
            "GET", "https://httpbin.org/user-agent",
            headers={"User-Agent": "Memorious Test"}
        )
        context_http_response = ContextHttpResponse(http, request)
        assert isinstance(context_http_response.raw, bytes)
        assert isinstance(context_http_response.text, str)
        assert context_http_response.json == {"user-agent": "Memorious Test"}

    def test_html(self, http):
        request = Request("GET", "https://httpbin.org/html")
        context_http_response = ContextHttpResponse(http, request)
        assert isinstance(context_http_response.html, html.HtmlElement)

    def test_xml(self, http):
        request = Request("GET", "https://httpbin.org/xml")
        context_http_response = ContextHttpResponse(http, request)
        assert isinstance(context_http_response.xml, etree._ElementTree)

    def test_apply_data(self, http):
        context_http_response = ContextHttpResponse(http)
        assert context_http_response.url is None
        assert context_http_response.status_code is None
        context_http_response.apply_data(data={
            "status_code": 200,
            "url": "https://httpbin.org/get"
        })
        assert context_http_response.url == "https://httpbin.org/get"
        assert context_http_response.status_code == 200

    def test_deserialize(self, http):
        data = {
            "status_code": 200,
            "url": "https://httpbin.org/get"
        }
        context_http_response = ContextHttpResponse.deserialize(http, data)
        assert isinstance(context_http_response, ContextHttpResponse)
        assert context_http_response.url == "https://httpbin.org/get"
        assert context_http_response.status_code == 200

    def test_close(self, http):
        request = Request("GET", "https://httpbin.org/get")
        context_http_response = ContextHttpResponse(http, request)
        file_path = context_http_response.fetch()
        assert os.path.exists(file_path)
        context_http_response.close()
        # assert not os.path.exists(file_path)

    @pytest.mark.parametrize("url,status_code", [
        ("https://httpbin.org/status/404", 404),
        ("https://httpbin.org/status/500", 500),
        ("https://httpbin.org/status/200", 200)
    ])
    def test_status_code(self, url, status_code, http):
        request = Request("GET", url)
        context_http_response = ContextHttpResponse(http, request)
        assert context_http_response.status_code == status_code



import os
import uuid

import pytest

from memorious import settings
from memorious.logic.manager import CrawlerManager
from memorious.logic.context import Context
from memorious.logic.http import ContextHttp


settings.TESTING = True


def get_crawler_dir():
    file_path = os.path.realpath(__file__)
    crawler_dir = os.path.normpath(os.path.join(
        file_path, "../testdata/config"
    ))
    return crawler_dir


@pytest.fixture(scope="module")
def crawler_dir():
    return get_crawler_dir()


def get_manager():
    manager = CrawlerManager()
    manager.load_path(get_crawler_dir())
    settings._manager = manager
    return manager


@pytest.fixture(scope="module")
def manager():
    return get_manager()


def get_crawler():
    return get_manager().get("occrp_web_site")


@pytest.fixture(scope="module")
def crawler():
    return get_crawler()


def get_stage():
    cr = get_crawler()
    return cr.get(cr.init_stage)


@pytest.fixture(scope="module")
def stage():
    return get_stage()


def get_context():
    ctx = Context(get_crawler(), get_stage(), {"foo": "bar"})
    ctx.run_id = str(uuid.uuid4())
    return ctx


@pytest.fixture(scope="module")
def context():
    return get_context()


@pytest.fixture(scope="module")
def http():
    return ContextHttp(get_context())



import logging
from servicelayer.worker import Worker

from memorious import settings
from memorious.logic.context import Context
from memorious.core import manager, conn, get_rate_limit

log = logging.getLogger(__name__)


class MemoriousWorker(Worker):
    def boot(self):
        self.scheduler = get_rate_limit('scheduler',
                                        unit=60,
                                        interval=settings.SCHEDULER_INTERVAL,
                                        limit=1)

    def periodic(self):
        if self.scheduler.check() and not settings.DEBUG:
            log.info("Running scheduled crawlers ...")
            self.scheduler.update()
            manager.run_scheduled()

    def handle(self, task):
        data = task.payload
        stage = task.stage.stage
        state = task.context
        context = Context.from_state(state, stage)
        if context.crawler.disabled:
            return
        context.execute(data)

    def after_task(self, task):
        if task.job.is_done():
            stage = task.stage.stage
            state = task.context
            context = Context.from_state(state, stage)
            context.crawler.aggregate(context)

    def get_stages(self):
        return list({str(stage) for _, stage in manager.stages})


def get_worker():
    return MemoriousWorker(conn=conn)



import os
import uuid
import shutil
import logging
from copy import deepcopy
from tempfile import mkdtemp
from contextlib import contextmanager
from servicelayer.cache import make_key
from servicelayer.util import load_json, dump_json

from memorious.core import manager, storage, conn
from memorious.core import datastore
from memorious.model import Event, Queue, Crawl
from memorious.logic.http import ContextHttp
from memorious.logic.check import ContextCheck
from memorious.util import random_filename
from memorious.exc import QueueTooBigError


class Context(object):
    """Provides state tracking and methods for operation interactions."""

    def __init__(self, crawler, stage, state):
        self.crawler = crawler
        self.stage = stage
        self.state = state
        self.params = stage.params
        self.incremental = state.get('incremental')
        self.run_id = state.get('run_id') or uuid.uuid1().hex
        self.work_path = mkdtemp()
        self.log = logging.getLogger('%s.%s' % (crawler.name, stage.name))
        self.http = ContextHttp(self)
        self.datastore = datastore
        self.check = ContextCheck(self)

    def get(self, name, default=None):
        """Get a configuration value and expand environment variables."""
        value = self.params.get(name, default)
        if isinstance(value, str):
            value = os.path.expandvars(value)
        return value

    def emit(self, rule='pass', stage=None, data={}, delay=None,
             optional=False):
        """Invoke the next stage, either based on a handling rule, or by calling
        the `pass` rule by default."""
        if stage is None:
            stage = self.stage.handlers.get(rule)
        if optional and stage is None:
            return
        if stage is None or stage not in self.crawler.stages:
            self.log.info("No next stage: %s (%s)" % (stage, rule))
            return
        state = self.dump_state()
        delay = delay or self.crawler.delay
        Queue.queue(stage, state, data)

    def recurse(self, data={}, delay=None):
        """Have a stage invoke itself with a modified set of arguments."""
        return self.emit(stage=self.stage.name,
                         data=data,
                         delay=delay)

    def execute(self, data):
        """Execute the crawler and create a database record of having done
        so."""
        if Crawl.is_aborted(self.crawler, self.run_id):
            return

        try:
            Crawl.operation_start(self.crawler, self.stage, self.run_id)
            self.log.info('[%s->%s(%s)]: %s',
                          self.crawler.name,
                          self.stage.name,
                          self.stage.method_name,
                          self.run_id)
            return self.stage.method(self, data)
        except QueueTooBigError as qtbe:
            self.emit_warning(str(qtbe))
        except Exception as exc:
            self.emit_exception(exc)
        finally:
            Crawl.operation_end(self.crawler, self.run_id)
            shutil.rmtree(self.work_path)

    def emit_warning(self, message, type=None, *args):
        if len(args):
            message = message % args
        self.log.warning(message)
        return Event.save(self.crawler,
                          self.stage,
                          Event.LEVEL_WARNING,
                          self.run_id,
                          error=type,
                          message=message)

    def emit_exception(self, exc):
        self.log.exception(exc)
        return Event.save(self.crawler,
                          self.stage,
                          Event.LEVEL_ERROR,
                          self.run_id,
                          error=exc.__class__.__name__,
                          message=str(exc))

    def set_tag(self, key, value):
        data = dump_json(value)
        key = make_key(self.crawler, "tag", key)
        return conn.set(key, data, ex=self.crawler.expire)

    def get_tag(self, key):
        value = conn.get(make_key(self.crawler, "tag", key))
        if value is not None:
            return load_json(value)

    def check_tag(self, key):
        return conn.exists(make_key(self.crawler, "tag", key))

    def skip_incremental(self, *criteria):
        """Perform an incremental check on a set of criteria.

        This can be used to execute a part of a crawler only once per an
        interval (which is specified by the ``expire`` setting). If the
        operation has already been performed (and should thus be skipped),
        this will return ``True``. If the operation needs to be executed,
        the returned value will be ``False``.
        """
        if not self.incremental:
            return False

        # this is pure convenience, and will probably backfire at some point.
        key = make_key(*criteria)
        if key is None:
            return False

        if self.check_tag(key):
            return True

        self.set_tag(key, None)
        return False

    def store_file(self, file_path, content_hash=None):
        """Put a file into permanent storage so it can be visible to other
        stages."""
        return storage.archive_file(file_path, content_hash=content_hash)

    def store_data(self, data, encoding='utf-8'):
        """Put the given content into a file, possibly encoding it as UTF-8
        in the process."""
        path = random_filename(self.work_path)
        try:
            with open(path, 'wb') as fh:
                if isinstance(data, str):
                    data = data.encode(encoding)
                if data is not None:
                    fh.write(data)
            return self.store_file(path)
        finally:
            try:
                os.unlink(path)
            except OSError:
                pass

    @contextmanager
    def load_file(self, content_hash, file_name=None, read_mode='rb'):
        file_path = storage.load_file(content_hash,
                                      file_name=file_name,
                                      temp_path=self.work_path)
        if file_path is None:
            yield None
        else:
            try:
                with open(file_path, mode=read_mode) as fh:
                    yield fh
            finally:
                storage.cleanup_file(content_hash,
                                     temp_path=self.work_path)

    def dump_state(self):
        state = deepcopy(self.state)
        state['crawler'] = self.crawler.name
        state['run_id'] = self.run_id
        return state

    @classmethod
    def from_state(cls, state, stage):
        state_crawler = state.get('crawler')
        crawler = manager.get(state_crawler)
        if crawler is None:
            raise RuntimeError("Missing crawler: [%s]" % state_crawler)
        stage = crawler.get(stage)
        if stage is None:
            raise RuntimeError('[%r] has no stage: %s' % (crawler, stage))
        return cls(crawler, stage, state)

    def __repr__(self):
        return '<Context(%r, %r)>' % (self.crawler, self.stage)



import logging

from servicelayer.jobs import Stage, Job, Dataset

from memorious.core import conn
from memorious.settings import MAX_QUEUE_LENGTH
from memorious.exc import QueueTooBigError

log = logging.getLogger(__name__)


class Queue(object):
    """Manage the execution of tasks in the system."""

    @classmethod
    def queue(cls, stage, state, data):
        crawler = state.get('crawler')
        job = Job(conn, str(crawler), state['run_id'])
        job_stage = job.get_stage(stage)
        queue_length = job_stage.get_status().get('pending')
        if queue_length > MAX_QUEUE_LENGTH:
            msg = "queue for %s:%s too big."
            raise QueueTooBigError(msg % (str(crawler), stage))
        job_stage.queue(payload=data, context=state)




from setuptools import setup, find_packages

setup(
    name='memorious',
    version='1.2.10',
    description="A minimalistic, recursive web crawling library for Python.",
    long_description="",
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    keywords='',
    author='Organized Crime and Corruption Reporting Project',
    author_email='data@occrp.org',
    url='http://github.com/alephdata/memorious',
    license='MIT',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    namespace_packages=[],
    include_package_data=True,
    package_data={
        'memorious': [
            'migrate/*',
            'migrate/versions/*',
            'migrate/ui/static/*',
            'migrate/ui/templates/*'
        ]
    },
    zip_safe=False,
    install_requires=[
        'banal >= 0.4.2',
        'click',
        'requests[security] >= 2.21.0',
        'requests_ftp',
        'lxml >= 3',
        'PyYAML >= 3.10',
        'normality >= 1.0.0',
        'tabulate',
        'dataset >= 1.0.8',
        'servicelayer[google,amazon] == 1.9.3',
        'urlnormalizer == 1.2.5',
        'pantomime == 0.3.3',
        'dateparser',
        'stringcase',
        'flask',
        'babel'
    ],
    entry_points={
        'console_scripts': [
            'memorious = memorious.cli:main'
        ],
        'memorious.operations': [
            'memorious = memorious.cli:main',
            'fetch = memorious.operations.fetch:fetch',
            'dav_index = memorious.operations.fetch:dav_index',
            'session = memorious.operations.fetch:session',
            'parse = memorious.operations.parse:parse',
            'clean_html = memorious.operations.clean:clean_html',
            'seed = memorious.operations.initializers:seed',
            'sequence = memorious.operations.initializers:sequence',
            'dates = memorious.operations.initializers:dates',
            'enumerate = memorious.operations.initializers:enumerate',
            'inspect = memorious.operations.debug:inspect',
            'documentcloud_query = memorious.operations.documentcloud:documentcloud_query',  # noqa
            'directory = memorious.operations.store:directory',
            'extract = memorious.operations.extract:extract',
            'db = memorious.operations.db:db',
            'ftp_fetch = memorious.operations.ftp:ftp_fetch',
        ]
    },
    extras_require={
        'dev': [
            'pytest',
            'pytest-env',
            'pytest-cov',
            'pytest-mock',
        ],
        'ocr': [
            'tesserocr',
        ],
    }
)



import os
from datetime import datetime
import requests
import requests_ftp


def ftp_fetch(context, data):
    url = data.get('url')
    context.log.info("FTP fetch: %s", url)
    requests_ftp.monkeypatch_session()
    session = requests.Session()
    username = context.get('username', 'Anonymous')
    password = context.get('password', 'anonymous@ftp')

    cached = context.get_tag(url)
    if cached is not None:
        context.emit(rule='pass', data=cached)
        return

    resp = session.retr(url, auth=(username, password))
    if resp.status_code < 399:
        data.update({
            'status_code': resp.status_code,
            'retrieved_at': datetime.utcnow().isoformat(),
            'content_hash': context.store_data(data=resp.content)
        })
        context.set_tag(url, data)
        context.emit(rule='pass', data=data)
    else:
        resp = session.nlst(url, auth=(username, password))
        for child in resp.iter_lines(decode_unicode=True):
            child_data = data.copy()
            child_data['url'] = os.path.join(url, child)
            # context.log.info("FTP directory child: %(url)s", child_data)
            context.emit(rule='child', data=child_data)



import os
import tempfile
from memorious.operations.extract import extract_7zip, extract_tar, extract_zip


def test_extract_7zip(context):
    file_path = os.path.realpath(__file__)
    archive_path = os.path.normpath(os.path.join(
        file_path, "../testdata/test.7z"
    ))
    extract_dir = tempfile.mkdtemp(prefix="memorious_test")
    assert extract_7zip(archive_path, extract_dir, context) == [
        os.path.join(extract_dir, "test/a/1.txt")
    ]


def test_extract_zip(context):
    file_path = os.path.realpath(__file__)
    archive_path = os.path.normpath(os.path.join(
        file_path, "../testdata/test.zip"
    ))
    extract_dir = tempfile.mkdtemp(prefix="memorious_test")
    assert extract_zip(archive_path, extract_dir, context) == [
        os.path.join(extract_dir, "test/a/1.txt")
    ]


def test_extract_tar(context):
    file_path = os.path.realpath(__file__)
    archive_path = os.path.normpath(os.path.join(
        file_path, "../testdata/test.tar.gz"
    ))
    extract_dir = tempfile.mkdtemp(prefix="memorious_test")
    assert extract_tar(archive_path, extract_dir, context) == [
        os.path.join(extract_dir, "test/a/1.txt")
    ]



import math
from urllib.parse import urlencode
import logging
from flask import Flask, request, redirect
from flask import render_template, abort, url_for
from babel.numbers import format_number
from babel.dates import format_date, format_datetime

from memorious.core import settings, manager, init_memorious
from memorious.model import Event

PAGE_SIZE = 50
app = Flask(__name__)
init_memorious()
if settings.DEBUG:
    logging.basicConfig(level=logging.DEBUG)
else:
    logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


@app.template_filter('number')
def number_filter(s, default=''):
    if s is None or s == 0 or not len(str(s)):
        return default
    return format_number(s, locale='en_GB')


@app.template_filter('datetime')
def datetime_filter(s):
    if s is None or s == 0 or not len(str(s)):
        return ''
    return format_datetime(s, locale='en_GB', format='short')


@app.template_filter('date')
def date_filter(s):
    if s is None or s == 0 or not len(str(s)):
        return ''
    return format_date(s, locale='en_GB', format='short')


def state_change(name, value):
    state = [(name, value)]
    for aname, avalue in request.args.items():
        if aname != name:
            state.append((aname, avalue))
    return '?' + urlencode(state)


@app.context_processor
def context():
    return {
        'version': settings.VERSION,
        'num_crawlers': len(manager),
        'state_change': state_change
    }


def get_crawler(name):
    crawler = manager.get(name)
    if crawler is None:
        abort(404)
    return crawler


def redirect_crawler(crawler):
    if request.form.get('return') == 'index':
        return redirect(url_for('.index'))
    return redirect(url_for('.crawler', name=crawler.name))


@app.route('/')
def index():
    """Generate a list of all crawlers, alphabetically, with op counts."""
    crawlers = []
    for crawler in manager:
        data = Event.get_counts(crawler)
        data['last_active'] = crawler.last_run
        data['total_ops'] = crawler.op_count
        data['running'] = crawler.is_running
        data['crawler'] = crawler
        crawlers.append(data)
    return render_template('index.html', crawlers=crawlers)


@app.route('/crawlers/<name>')
def crawler(name):
    crawler = get_crawler(name)
    stages = []
    for stage in crawler:
        data = Event.get_stage_counts(crawler, stage)
        data['total_ops'] = stage.op_count
        data['stage'] = stage
        stages.append(data)
    runs = list(crawler.runs)
    for run in runs:
        run.update(Event.get_run_counts(crawler, run['run_id']))
    runs = sorted(runs, key=lambda r: r.get('start'), reverse=True)
    return render_template('crawler.html',
                           crawler=crawler,
                           stages=stages,
                           runs=runs)


@app.route('/crawlers/<name>/events')
def events(name):
    crawler = get_crawler(name)
    page = int(request.args.get('page', 1))
    start = (max(1, page) - 1) * PAGE_SIZE
    end = start + PAGE_SIZE
    run_id = request.args.get('run_id')
    level = request.args.get('level')
    stage_name = request.args.get('stage_name')

    if stage_name:
        events = Event.get_stage_events(crawler, stage_name, start, end, level)
    elif run_id:
        events = Event.get_run_events(crawler, run_id, start, end, level)
    else:
        events = Event.get_crawler_events(crawler, start, end, level)
    total = len(events)
    pages = int(math.ceil((float(total) / PAGE_SIZE)))
    return render_template('events.html',
                           crawler=crawler,
                           results=events,
                           page=page,
                           pages=pages)


@app.route('/crawlers/<name>/config')
def config(name):
    crawler = get_crawler(name)
    return render_template('config.html', crawler=crawler)


@app.route('/invoke/<crawler>/run', methods=['POST'])
def crawler_run(crawler):
    crawler = get_crawler(crawler)
    crawler.run()
    return redirect_crawler(crawler)


@app.route('/invoke/<crawler>/cancel', methods=['POST'])
def crawler_cancel(crawler):
    crawler = get_crawler(crawler)
    crawler.cancel()
    return redirect_crawler(crawler)


@app.route('/invoke/<crawler>/flush', methods=['POST'])
def crawler_flush(crawler):
    crawler = get_crawler(crawler)
    crawler.flush()
    return redirect_crawler(crawler)


@app.route('/invoke/<crawler>/flush-events', methods=['POST'])
def crawler_flush_events(crawler):
    crawler = get_crawler(crawler)
    crawler.flush_events()
    return redirect_crawler(crawler)


if __name__ == '__main__':
    app.run(host='0.0.0.0', port=4000, debug=True)
